
//		FindMyWindow.cpp


#ifndef FIND_MY_WINDOW_VER
#define FIND_MY_WINDOW_VER				1.0.7.0


#define MY_WINDOW_UNKNOWN				0
#define MY_WINDOW_CONSOLE				1
#define MY_WINDOW_TAKE_COMMAND			2
#define MY_WINDOW_CONSOLE2				3
#define MY_WINDOW_CONEMU				4
#define MY_WINDOW_WINDOWS_TERMINAL		16


#define MAX_CONSOLE2_WINDOWS			32
#define MY_WINDOW_MAX_WINDOWS			32



struct _FMW_FindWinData {
	LPCTSTR ClassName;
	unsigned int MaxWindows;
	unsigned int n;
	HWND *hWnd;
	DWORD *Pid;
};


struct _FMW_FindChildWinData {
	LPCTSTR ClassName;
	HWND hWnd;
};


struct _FMW_FindTabWinData {
	LPTSTR ConsoleTitle;
	HWND hWnd;
};



BOOL CALLBACK FMW_FindProc$( HWND hWnd, LPARAM lParam )
{
	wchar_t Class[SHORT_BUF_LEN]			= L"";
	_FMW_FindWinData *Data					= (_FMW_FindWinData*) lParam;


	if ( !RealGetWindowClass( hWnd, Class, SHORT_BUF_LEN ) )
		return TRUE;

	if ( wcscmp( Class, Data->ClassName ) )
		return TRUE;

	if ( Data->n >= Data->MaxWindows )
		return FALSE;

	Data->hWnd[Data->n] = hWnd;

	if ( Data->Pid )
		GetWindowThreadProcessId( hWnd, Data->Pid + Data->n );

	Data->n += 1;
	return TRUE;
}



int FMW_FindWindowsByClass( LPCTSTR ClassName, HWND *hWnd, DWORD *Pid, unsigned int MaxWindows )
{
	if ( !ClassName || !hWnd || !MaxWindows )
		return -1;
	else if ( ClassName[0] == '\0' )
		return -1;


	_FMW_FindWinData Data;


	memset( hWnd, 0x00, MaxWindows * sizeof( HWND ) );
	if ( Pid )
		memset( Pid, 0x00, MaxWindows * sizeof( DWORD ) );

	Data.ClassName	= ClassName;
	Data.MaxWindows	= MaxWindows;
	Data.n			= 0;
	Data.hWnd		= hWnd;
	Data.Pid		= Pid;


	EnumWindows( FMW_FindProc$, (LPARAM) &Data );
	return Data.n;
}



BOOL FMW_Offscreen( HWND hWnd )
{
	if ( !hWnd )
		return FALSE;


	HMONITOR Monitor		= MonitorFromWindow( hWnd, MONITOR_DEFAULTTONULL );
	return ( Monitor == NULL );
}



HWND FMW_FindConsole2()
{
	HWND hWnd[MY_WINDOW_MAX_WINDOWS];
	DWORD Pid[MY_WINDOW_MAX_WINDOWS];

	int NumFound				= 0;
	DWORD ThisPid				= GetCurrentProcessId();


	NumFound = FMW_FindWindowsByClass( L"Console_2_Main", hWnd, Pid, MY_WINDOW_MAX_WINDOWS );
	if ( NumFound < 1 )
		return NULL;


	while ( ThisPid && ( ThisPid != 0xffffffff ) ) {
		for ( int i = 0; i < NumFound; i++ )
			if ( ThisPid == Pid[i] )
				return hWnd[i];

		ThisPid = GetParentPID( ThisPid );
	};

	return NULL;
}



HWND FMW_FindConEmu()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";

	HWND hWnd			= NULL;


	GetEnvironmentVariable( L"ConEmuHWND", temp, SHORT_BUF_LEN );

	if ( iswdigit( temp[0] ) ) {
		ParseInt( temp, (unsigned int*) &hWnd, NULL );

		if ( hWnd ) {
			temp[0] = '\0';
			RealGetWindowClass( hWnd, temp, SHORT_BUF_LEN );
			
			if ( !wcscmp( temp, L"VirtualConsoleClass" ) )
				return hWnd;
		}
	}

	return NULL;
}



HWND FMW_ConEmuDrawHwnd()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";

	HWND hWnd			= NULL;


	GetEnvironmentVariable( L"ConEmuDrawHWND", temp, SHORT_BUF_LEN );

	if ( iswdigit( temp[0] ) ) {
		ParseInt( temp, (unsigned int*) &hWnd, NULL );

		if ( hWnd ) {
			temp[0] = '\0';
			RealGetWindowClass( hWnd, temp, SHORT_BUF_LEN );

			if ( !wcscmp( temp, L"VirtualConsoleClass" ) )
				return hWnd;
		}
	}

	return NULL;
}



BOOL CALLBACK FMW_FindChildWinProc$( HWND hWnd, LPARAM lParam )
{
	wchar_t Class[SHORT_BUF_LEN]			= L"";
	_FMW_FindChildWinData *Data				= (_FMW_FindChildWinData*) lParam;


	if ( !RealGetWindowClass( hWnd, Class, SHORT_BUF_LEN ) )
		return TRUE;

	if ( wcscmp( Class, Data->ClassName ) )
		return TRUE;

	Data->hWnd = hWnd;
	return FALSE;
}



HWND FMW_FindChildWindowByClass( HWND hWnd, LPCTSTR ClassName )
{
	if ( !hWnd || !ClassName )
		return NULL;
	else if ( ClassName[0] == '\0' )
		return NULL;

	if ( !IsWindow( hWnd ) )
		return NULL;


	_FMW_FindChildWinData Data;


	Data.ClassName	= ClassName;
	Data.hWnd		= NULL;

	EnumChildWindows( hWnd, FMW_FindChildWinProc$, (LPARAM) &Data );
	return Data.hWnd;
}



HWND Console2ViewHwnd( HWND hWnd )
{
	return FMW_FindChildWindowByClass( hWnd, L"Console_2_View" );
}



/*

HWND FMW_WindowsTerminalViewHwnd( HWND hWnd )
{
	return FMW_FindChildWindowByClass( hWnd, L"Windows.UI.Composition.DesktopWindowContentBridge" );
}

*/



HWND FMW_FindWindowsTerminal()
{
	wchar_t temp[SHORT_BUF_LEN]		= L"";

	HWND hWnd[MY_WINDOW_MAX_WINDOWS];
	DWORD Pid[MY_WINDOW_MAX_WINDOWS];

	int NumFound				= 0;
	DWORD ThisPid				= GetCurrentProcessId();


	/*
	GetEnvironmentVariable( L"WT_SESSION", temp, SHORT_BUF_LEN );
	if ( temp[0] == '\0' )
		return NULL;

	temp[0] = '\0';
	*/

	RealGetWindowClass( GetConsoleWindow(), temp, SHORT_BUF_LEN );
	if ( _wcsicmp( temp, L"PseudoConsoleWindow" ) )
		return 0;

	NumFound = FMW_FindWindowsByClass( L"CASCADIA_HOSTING_WINDOW_CLASS", hWnd, Pid, MY_WINDOW_MAX_WINDOWS );
	if ( NumFound < 1 )
		return NULL;


	while ( ThisPid && ( ThisPid != 0xffffffff ) ) {
		for ( int i = 0; i < NumFound; i++ )
			if ( ThisPid == Pid[i] )
				return hWnd[i];

		ThisPid = GetParentPID( ThisPid );
	};

	return NULL;
}



int FMW_DummyConsoleOutput( unsigned int n )
{
	for ( unsigned int i = 0; i < n; i++ )
		_cputws( L"A\bB\bC\bD\b \b" );

	return 0;
}



HWND FMW_AltFindTakeCommand()
{
	HWND hWnd[MY_WINDOW_MAX_WINDOWS];
	DWORD Pid[MY_WINDOW_MAX_WINDOWS];

	int NumFound				= 0;

	DWORD BufferSize			= 64;
	DWORD ProcessCount			= BufferSize + 1;
	DWORD *ProcessListBuffer	= NULL;


	NumFound = FMW_FindWindowsByClass( L"XTPMainFrame", hWnd, Pid, MY_WINDOW_MAX_WINDOWS );
	if ( NumFound < 1 )
		return NULL;


	FMW_DummyConsoleOutput( 256 );
	tty_yield( 0 );

	while ( ProcessCount > BufferSize ) {
		if ( ProcessListBuffer ) {
			free( ProcessListBuffer );
			ProcessListBuffer = NULL;
		}

		ProcessListBuffer = (DWORD*) malloc( BufferSize * sizeof( DWORD ) );
		if ( !ProcessListBuffer )
			return NULL;

		ProcessCount = GetConsoleProcessList( ProcessListBuffer, BufferSize );
	}

	if ( !ProcessCount ) {
		free( ProcessListBuffer );
		return NULL;
	}


	for ( unsigned int i = 0; i < ProcessCount; i++ )
		if ( ProcessListBuffer[i] )
			for ( int j = 0; j < NumFound; j++ )
				if ( ProcessListBuffer[i] == Pid[j] )
					return hWnd[j];

	return NULL;
}



BOOL CALLBACK FMW_FindTabWinProc$( HWND hWnd, LPARAM lParam )
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";

	_FMW_FindTabWinData *Data			= (_FMW_FindTabWinData*) lParam;


	RealGetWindowClass( hWnd, temp, SHORT_BUF_LEN );
	if ( wcsncmp( temp, L"Afx:", 4 ) )
		return TRUE;

	temp[0] = '\0';
	GetWindowText( hWnd, temp, SHORT_BUF_LEN );
	if ( wcscmp( temp, Data->ConsoleTitle ) )
		return TRUE;

	Data->hWnd = hWnd;
	return FALSE;
}



HWND FMW_TakeCommandTabHwnd( HWND TakeCmdWindow )
{
	wchar_t temp[SHORT_BUF_LEN]		= L"";

	HWND hwnd				= NULL;
	int rv					= 0;


	rv = TakeCommandIPC( L"HVIEW", temp );

	if ( iswdigit( temp[0] ) ) {
		hwnd = (HWND) wcstoul( temp, NULL, 10 );

		if ( hwnd )
			if ( IsWindow( hwnd ) )
				return hwnd;
	}

	if ( ( rv == -1 ) && IsWindow( TakeCmdWindow ) ) {
		wchar_t OldTitle[SHORT_BUF_LEN]			= L"";
		wchar_t NewTitle[SHORT_BUF_LEN]			= L"";

		HWND ConsoleWindow			= GetConsoleWindow();
		_FMW_FindTabWinData Data;


		GetWindowText( ConsoleWindow, OldTitle, SHORT_BUF_LEN );
		wsprintf( NewTitle, L"\u00a4 %04X-%04X", GetCurrentProcessId(), rand() );
		SetWindowText( ConsoleWindow, NewTitle );

		Data.ConsoleTitle	= NewTitle;
		Data.hWnd			= NULL;

		for ( int i = 0; i < 20; i++ ) {
			EnumChildWindows( TakeCmdWindow, FMW_FindTabWinProc$, (LPARAM) &Data );
			if ( Data.hWnd )
				break;

			Sleep( 50 );
		}

		SetWindowText( ConsoleWindow, OldTitle );

		if ( Data.hWnd )
			return Data.hWnd;
	}

	return NULL;
}



HWND FindMyWindow( unsigned int *Type, HWND *View )
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";
	HWND hwnd							= NULL;
	HWND vhwnd							= NULL;
	int ipcrv							= 0;
	unsigned int t						= 0;


	//		First look for Take Command :

	ipcrv = TakeCommandIPC( L"HWND", temp );

	if ( iswdigit( temp[0] ) ) {
		hwnd = (HWND) wcstoul( temp, NULL, 10 );

		if ( hwnd ) {
			if ( IsWindow( hwnd ) ) {
				if ( Type )
					*Type = MY_WINDOW_TAKE_COMMAND;

				if ( View ) {
					vhwnd = FMW_TakeCommandTabHwnd( hwnd );

					if ( vhwnd )
						*View = vhwnd;
					else
						*View = hwnd;
				}

				return hwnd;
			}
		}
	}


	//		Then look for a regular console window :

	hwnd = GetConsoleWindow();

	if ( hwnd ) {
		if ( IsWindowVisible( hwnd ) ) {
			RealGetWindowClass( hwnd, temp, SHORT_BUF_LEN );
			if ( !wcscmp( temp, L"ConsoleWindowClass" ) ) {
				t = MY_WINDOW_CONSOLE;

				if ( FMW_Offscreen( hwnd ) )
					if ( vhwnd = FMW_FindConsole2() )
						hwnd = vhwnd, t = MY_WINDOW_CONSOLE2;

				if ( Type )
					*Type = t;

				if ( View )
					*View = hwnd;

				return hwnd;
			}
		}
	}


	//		Next look for Window Terminal :

	hwnd = FMW_FindWindowsTerminal();

	if ( hwnd ) {
		if ( Type )
			*Type = MY_WINDOW_WINDOWS_TERMINAL;

		/*
		if ( View ) {
			vhwnd = FMW_WindowsTerminalViewHwnd( hwnd );

			if ( vhwnd )
				*View = vhwnd;
			else
				*View = hwnd;
		}
		*/

		if ( View )
			*View = hwnd;

		return hwnd;
	}


	//		Look for ConEmu :

	hwnd = FMW_FindConEmu();

	if ( hwnd ) {
		if ( Type )
			*Type = MY_WINDOW_CONEMU;

		if ( View ) {
			vhwnd = FMW_ConEmuDrawHwnd();

			if ( vhwnd )
				*View = vhwnd;
			else
				*View = hwnd;
		}

		return hwnd;
	}


	//		Finally check for Console2 :

	hwnd = FMW_FindConsole2();
	if ( hwnd ) {
		if ( Type )
			*Type = MY_WINDOW_CONSOLE2;

		if ( View ) {
			vhwnd = Console2ViewHwnd( hwnd );

			if ( vhwnd )
				*View = vhwnd;
			else
				*View = hwnd;
		}

		return hwnd;
	}


	//		If TakeCommandIPC() failed, look for Take Command another way :

	hwnd = FMW_AltFindTakeCommand();
	if ( hwnd ) {
		if ( Type )
			*Type = MY_WINDOW_TAKE_COMMAND;

		if ( View ) {
			vhwnd = FMW_TakeCommandTabHwnd( hwnd );

			if ( vhwnd )
				*View = vhwnd;
			else
				*View = hwnd;
		}

		return hwnd;
	}


	//		Give up!

	if ( Type )
		*Type = MY_WINDOW_UNKNOWN;

	if ( View )
		*View = NULL;

	return NULL;
}



HWND FindMyWindow()
{
	return FindMyWindow( NULL, NULL );
}



HWND FindTakeCmdWindow()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";
	HWND hwnd							= NULL;
	int ipcrv							= 0;


	//		First try the IPC interface :

	ipcrv = TakeCommandIPC( L"HWND", temp );

	if ( iswdigit( temp[0] ) ) {
		hwnd = (HWND) wcstoul( temp, NULL, 10 );

		if ( hwnd )
			if ( IsWindow( hwnd ) )
				return hwnd;
	}


	//		If TakeCommandIPC() failed, look for Take Command another way :

	hwnd = FMW_AltFindTakeCommand();
	if ( hwnd )
		return hwnd;


	//		Then give up.

	return NULL;
}


#endif	//	FIND_MY_WINDOW_VER


/*

	1.0.0.0		2022-07-14

	1.0.1.0		2022-07-15
				Tweak to accomodate a newer build of Windows Terminal:  Check that the
				console window's class is really "ConsoleWindowClass".

	1.0.2.0		2022-07-15
				Hacked in support for ConEmu.

	1.0.3.0		2022-08-11
				Added a third argument to report the handle of the view window, if
				available.

	1.0.4.0		2022-08-29
				Removed FMW_WindowsTerminalViewHwnd(), which so far as I can tell is
				perfectly useless.

	1.0.5.0		2023-12-04
				Added FindTakeCmdWindow() for functions which only care about Take
				Command, not ConEmu, Windows Terminal, etc.  Added a tty_yield() to
				FMW_AltFindTakeCommand().

	1.0.6.0		2023-12-05
				Console2 can sometimes attach console windows which are visible but
				offscreen.  Added a check for this situation.

	1.0.7.0		2024-04-01
				FMW_FindWindowsTerminal() no longer relies on the WT_SESSION
				environment variable.  Instead, it checks that
				RealGetWindowClass( GetConsoleWindow()) returns a class name of
				"PseudoConsoleWindow".

*/
