
//		SetThemeCmd.cpp  --  implements the SETTHEME command


#ifndef SETTHEME_CMD
#define SETTHEME_CMD				1.0.0.1


#include "ForceFGWin.cpp"



const wchar_t SetThemeHelpText[] = L"Apply a Windows theme file.\n\
\n\
SETTHEME filename\r\n\n\
The filename is required.\n  ";



HWND SetTheme_FindWindowByTwoClasses( LPTSTR Class1, LPTSTR Class2 )
{
	if ( !Class1 || !Class2 )
		return NULL;
	else if ( ( Class1[0] == '\0' ) || ( Class2[0] == '\0' ) )
		return NULL;


	HWND rv = FindWindow( Class1, NULL );

	if ( rv == NULL )
		rv = FindWindow( Class2, NULL );


	return rv;
}



DLLExports int WINAPI settheme( LPTSTR inString )
{
	if ( !inString )
		return -1;


	wchar_t Filename[MAX_FILENAME_LEN]		= L"";
	unsigned int Seconds					= 8;
	BOOL Quiet								= FALSE;
	BOOL Debug								= FALSE;

	LPTSTR Ext								= NULL;
	HWND hWnd								= NULL;
	HWND FgWindow							= NULL;
	unsigned int Count						= 0;
	BOOL Found								= FALSE;
	
	wchar_t WinClass1[]						= L"CabinetWClass";
	wchar_t WinClass2[]						= L"ApplicationFrameWindow";


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) )
				return ShowCmdHelp( SetThemeHelpText, ( arg[i][2] == '?' ) );
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Quiet = TRUE;
			else if ( !_wcsicmp( arg[i], L"/$" ) )
				Debug = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/T", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &Seconds, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad time: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( ( Seconds < 1 ) || ( Seconds > 30 ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Time must be 1 - 30: %u\r\n", Seconds );
					return 1;
				}
			}
			else if ( arg[i][0] == '/' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( Filename[0] == '\0' )
				QueryTrueName( arg[i], Filename );
			else {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}
		}
	}


	if ( Filename[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return 1;
	}


	if ( _ExtensionPart( Filename, NULL ) == NULL )
		wcscat_s( Filename, MAX_FILENAME_LEN, L".theme" );

	Ext = _ExtensionPart( Filename, NULL );

	if ( Ext )
		if ( _wcsicmp( Ext, L".theme" ) )
			Ext = NULL;

	if ( !Ext ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Extension must be .THEME: \"%s\"\r\n", Filename );
		return 1;
	}
	
	if ( !IsAFile( Filename ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"File not found: \"%s\"\r\n", Filename );
		return 2;
	}


	FgWindow = GetForegroundWindow();

	if ( Debug )
		_cputws( L"  *  Close any open Themes windows:\r\n" );


	do {
		hWnd = SetTheme_FindWindowByTwoClasses( WinClass1, WinClass2 );

		if ( hWnd ) {
			SendMessage( hWnd, WM_CLOSE, 0, NULL );
			Found = TRUE;
		}

	} while ( hWnd );

	if ( Found )
		Sleep( 200 );


	if ( Debug )
		_cputws( L"  *  Call ShellExecute() :\r\n" );

	ShellExecute( FgWindow, NULL, Filename, NULL, NULL, SW_MINIMIZE );

	do {
		hWnd = SetTheme_FindWindowByTwoClasses( WinClass1, WinClass2 );
		Count++;

		if ( !hWnd )
			Sleep( 10 );
	}
	while ( !hWnd && ( Count < ( Seconds * 100 ) ) );

	if ( !hWnd ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Personalization window not opening\r\n" );
		return 2;
	}


	if ( Debug )
		_cputws( L"  *  Move Themes window offscreen:\r\n" );

	MoveWindow( hWnd, -32000, -32000, 1200, 1000, FALSE );
	Sleep( 25 );
	MoveWindow( hWnd, -32000, -32000, 1200, 1000, FALSE );
	Sleep( 50 );
	MoveWindow( hWnd, -32000, -32000, 1200, 1000, FALSE );
	Sleep( 50 );
	MoveWindow( hWnd, -32000, -32000, 1200, 1000, FALSE );



	if ( !Quiet )
		_cputws( L"Wait: " );

	Sleep( 5000 );

	SendMessage( hWnd, WM_CLOSE, 0, NULL );

	if ( !Quiet )
		_cputws( L"\r      \r" );
	
	ForceForegroundWindow( FgWindow );
	return 0;
}


#endif	//	SETTHEME_CMD


/*

	1.0.0.0		2024-01-04

	1.0.0.1		2024-03-13
				#include "ForceFGWin.cpp"

*/
